//==============================================================================
//
//		Copyright 2009 IPK Gatersleben, AG Plant Bioinformatics
//		Developed by K. Pucknat, H. Rohn
//
//		Licensed under the Apache License, Version 2.0 (the "License");
//		you may not use this file except in compliance with the License.
//		You may obtain a copy of the License at
//
//			http://www.apache.org/licenses/LICENSE-2.0
//
//		Unless required by applicable law or agreed to in writing, software
//		distributed under the License is distributed on an "AS IS" BASIS,
//		WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//		See the License for the specific language governing permissions and
//		limitations under the License. 
//
//		Portions of this software is based in part on the CyberVRML97 package written by Satoshi Konno.
//
//==============================================================================

package anaglyphcanvas3d;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.MediaTracker;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ContainerEvent;
import java.awt.event.ContainerListener;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.util.Timer;
import java.util.TimerTask;

import javax.media.j3d.Canvas3D;
import javax.media.j3d.Screen3D;
import javax.media.j3d.View;
import javax.swing.JComponent;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;
import javax.vecmath.Point3d;


// TODO: Auto-generated Javadoc
// The Class AnaglyphCanvas3D.
/** 
 * Die Klasse <code>AnaglyphCanvas3D</code> ist eine Erweiterung der Klasse <code>Canvas3D</code>, 
 * welche einerseits das Rendern ohne Anaglyphen realisiert, als auch die F�rbung und Vereinigung der Halbbilder durchf�hrt.
 * @see Canvas3D
 */
public class AnaglyphCanvas3D extends Canvas3D {
	
	/** The Constant serialVersionUID. */
	private static final long serialVersionUID = 1L;
	
	//The int buffer image1. 
	/** 
	 * Array zur Speicherung der einzelnen Pixel des ersten gerenderten Halbbildes aus der Klasse <code>OffScreenCanvas3D</code>
	 * */
	private int[] intBufferImage1;
	
	//The int buffer image2. 
	/** 
	 * Array zur Speicherung der einzelnen Pixel des zweiten gerenderten Halbbildes aus der Klasse <code>OffScreenCanvas3D</code>
	 * */
	private int[] intBufferImage2;
	
	//The merged image. 
	/** 
	 * <code>BufferedImage</code> in welchem die beiden Halbbilder vereinigt werden
	 * */
	private BufferedImage mergedImage;

	//The stereo mode.
	/** 
	 * Enumeration zur Auswahl des Stereomodus
	 *  */
	private StereoMode stereoMode;
	
	//The anaglyph mode. 
	/** 
	 * Enumeration zur Auswahl des Anaglyphenmodus
	 * */
	private AnaglyphMode anaglyphMode;
	
	//The parent. 
	/** 
	 * <code>JComponent</code>, auf die der <code>AnaglyphCanvas3D</code> hinzugef�gt werden soll
	 * */
	private JComponent parent;

	// The canvas3 d offline image1. 
	/** 
	 * Erster <code>OffscreenCanvas3D</code>, der das erste Halbbild rendert
	 * */
	private OffScreenCanvas3D canvas3DOfflineImage1;
	
	//The canvas3 d offline image2. 
	/**
	 * Zweiter <code>OffscreenCanvas3D</code>, der das zweite Halbbild rendert
	 * */
	private OffScreenCanvas3D canvas3DOfflineImage2;
	
	// The red. 
	/**
	 * Variable f�r die Gewichtung des Rotkanal in der Graustufenberechnung
	 * @see AnaglyphMode 
	 * */
	private double red = 0.299;
	
	//The green. 
	/** 
	 * Variable f�r die Gewichtung des Gr�nkanal in der Graustufenberechnung
	 * @see AnaglyphMode 
	 * */
	private double green = 0.587;
	
	// The blue. 
	/**
	 * Variable f�r die Gewichtung des Blaukanal in der Graustufenberechnung
	 * @see AnaglyphMode 
	 * */
	private double blue = 0.114;

	private int transparencySort;
	private Dimension oldSize = new Dimension(0, 0);

	private Timer refreshSizeTimer;

	
	//Instantiates a new anaglyph canvas3 d.
	/**
	 * 
	 * Initializes a new <code>AnaglyphCanvas3D</code> object, which should functions as the usual <code>Canvas3D</code>
	 * and provides optional stereoscopic functionality optional.<br>
	 * The parent {@link JComponent} has to have a size (should be already added to other {@link JComponent}s, otherwise you won't see anything.<br>
	 * Don't forget to call the <code>close()</code> method, when closing the Canvas, otherwise a memory leak may occur.
	 * 
	 * 
	 * @param graphicsConf from e.g. <code>SimpleUniverse.getPreferredConfiguration()</code>	
	 * @param parent <code>JComponent</code>, to which the <code>AnaglyphCanvas3D</code> shall be added to.
	 * @see GraphicsConfiguration
	 * @see JComponent
	 */
	public AnaglyphCanvas3D(GraphicsConfiguration graphicsConf, JComponent parent) {
		super(graphicsConf, false);
		this.parent = parent;
		this.stereoMode = StereoMode.OFF;		
		this.anaglyphMode = AnaglyphMode.GRAY_ANAGLYPHS; 
	}			
	
	//Start anaglyph rendering.
	/**
	 * 
	 * Erzeugt 2 <code>OffScreenCanvas3D</code> Objekte,je einen pro Halbbild.
	 * 
	 */
	private void initializeAnaglyphRendering() {

		canvas3DOfflineImage1 = new OffScreenCanvas3D(getGraphicsConfiguration(), View.LEFT_EYE_VIEW);
		canvas3DOfflineImage2 = new OffScreenCanvas3D(getGraphicsConfiguration(), View.RIGHT_EYE_VIEW);
			
		View viewLeft = new View();
		View viewRight = new View();
	
		viewLeft.setPhysicalBody(getView().getPhysicalBody());	    				//Physical Body der linken Sicht mit dem Physikal Body der on-screen Sicht setzten
		viewLeft.setPhysicalEnvironment(getView().getPhysicalEnvironment());		//Physical Environment der linken Sicht mit der Physical Environment der on-screen Sicht setzten
	    viewRight.setPhysicalBody(getView().getPhysicalBody());						//Physical Body der rechten Sicht mit dem Physikal Body der on-screen Sicht setzten
        viewRight.setPhysicalEnvironment(getView().getPhysicalEnvironment());		//Physical Environment der rechten Sicht mit der Physical Environment der on-screen Sicht setzten
		
    	canvas3DOfflineImage1.setCanvas3DOnlineRef(this);							//erstem OffscreenCanvas3D Objekt die Referenz auf das AnaglyphCanvas3D Objekt �bergeben
        canvas3DOfflineImage2.setCanvas3DOnlineRef(this);							//zweitem OffscreenCanvas3D Objekt die Referenz auf das AnaglyphCanvas3D Objekt �bergeben

		Dimension size = new Dimension(parent.getWidth(), parent.getHeight());		//neue Dimension erzeugen mit der Breite und H�he f�r die zu rendernden Halbbilder
		canvas3DOfflineImage1.setOffScreenBuffer(size.width,size.height);			//Breite und H�he f�r das erste zu rendernde Halbbild setzten
		canvas3DOfflineImage2.setOffScreenBuffer(size.width,size.height);			//Breite und H�he f�r das zweite zu rendernde Halbbild setzten  
		
			
		viewLeft.attachViewPlatform(getView().getViewPlatform());					//Viewplatform des on-screen Renderers mit der linke Sicht verbinden
		viewRight.attachViewPlatform(getView().getViewPlatform());					//Viewplatform des on-screen Renderers mit der rechte Sicht verbinden
		
		viewLeft.addCanvas3D(canvas3DOfflineImage1);								//erste OffscreenCanvas3D Objekt der linken Sicht hinzuf�gen
		viewRight.addCanvas3D(canvas3DOfflineImage2);								//zweite OffscreenCanvas3D Objekt der rechten Sicht hinzuf�gen

		canvas3DOfflineImage1.getView().setTransparencySortingPolicy(transparencySort);
        canvas3DOfflineImage2.getView().setTransparencySortingPolicy(transparencySort);
				
		Screen3D s3D_On = this.getScreen3D();										//Screen3D von dem on-screen Renderer holen und in der s3D_On Variable speichern
		Screen3D s3D_OffLeft = canvas3DOfflineImage1.getScreen3D();					//Screen3D von dem ersten OffscreenCanvas3D Objekt holen und in der s3D_OffLeft Variable speichern 
		Screen3D s3D_OffRight = canvas3DOfflineImage2.getScreen3D();				//Screen3D von dem zweiten OffscreenCanvas3D Objekt holen und in der s3D_OffRight Variable speichern		
		s3D_OffLeft.setSize(s3D_On.getSize());										//setzt die Gr��e (in Pixeln) des ersten OffscreenCanvas3D Objekt mit der Gr��e des on-screen Renderers  
		s3D_OffRight.setSize(s3D_On.getSize());									    //setzt die Gr��e (in Pixeln) des zweiten OffscreenCanvas3D Objekt mit der Gr��e des on-screen Renderers
		s3D_OffLeft.setPhysicalScreenWidth(s3D_On.getPhysicalScreenWidth());		//setzt die PhysicalScreenWidth (in Metern) des ersten OffscreenCanvas3D Objekt mit der Breite des on-screen Renderers
		s3D_OffRight.setPhysicalScreenWidth(s3D_On.getPhysicalScreenWidth());		//setzt die PhysicalScreenWidth (in Metern) des zweiten OffscreenCanvas3D Objekt mit der Breite des on-screen Renderers
		s3D_OffLeft.setPhysicalScreenHeight(s3D_On.getPhysicalScreenHeight());		//setzt die PhysicalScreenHeight (in Metern) des ersten OffscreenCanvas3D Objekt mit der H�he des on-screen Renderers
		s3D_OffRight.setPhysicalScreenHeight(s3D_On.getPhysicalScreenHeight());		//setzt die PhysicalScreenHeight (in Metern) des zweiten OffscreenCanvas3D Objekt mit der H�he des on-screen Renderers	
		
		//on-screen Renderer muss das oberste Objekt auf der JComponent sein, damit die Behavior funktionieren
		parent.remove(this);														//on-screen Canvas3D vom JComponent entfernen		
		parent.add(canvas3DOfflineImage1, BorderLayout.CENTER);						//JComponent das erste OffscreenCanvas3D Objekt hinzuf�gen
		parent.add(canvas3DOfflineImage2, BorderLayout.CENTER);						//JComponent das zweiten OffscreenCanvas3D Objekt hinzuf�gen		
		parent.add(this, BorderLayout.CENTER);										//JComponent das AnaglyphCanvas3D Objekt hinzuf�gen
		//		
								
	}	

	//Sets the image1.
	/**
	 * Das erste Halbbild wird in seine Pixel aufgeteilt, um es sp�ter einzeln f�rben zu k�nnen.
	 * Eine F�rbemethode wird ensprechend des ausgew�hlten Anaglyphenmodus aufgrufen und ein Neuzeichnen des Bildes veranlasst. 
	 * @param image1 erste Halbbild
	 */
	protected void setImage1(BufferedImage image1){

		int width = image1.getWidth();
		int height = image1.getHeight();
			
		mergedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);		
  		int[] mergedIntBuffer = ( (DataBufferInt)mergedImage.getRaster().getDataBuffer() ).getData();	  		
  		intBufferImage1 = ((DataBufferInt)image1.getRaster().getDataBuffer()).getData(); 
  		 		
  		if(intBufferImage1 != null && intBufferImage2 != null){

			if(stereoMode == StereoMode.ANAGLYPH){
				switch(anaglyphMode){
				case REDBLUE_ANAGLYPHS:		redBlueAnaglyps(width, height, mergedIntBuffer);
											break;
				case REDGREEN_ANAGLYPHS:	redGreenAnaglyps(width, height, mergedIntBuffer);
											break;
				case GRAY_ANAGLYPHS:		grayAnaglyps(width, height, mergedIntBuffer);
											break;
				case COLOR_ANAGLYPHS:		colorAnaglyps(width, height, mergedIntBuffer);
											break;
				case HALFCOLOR_ANAGLYPHS:	halfColorAnaglyps(width, height, mergedIntBuffer);
											break;
				case OPTIMIZED_ANAGLYPHS:	optimizedAnaglyps(width, height, mergedIntBuffer);
											break;		
				}			
			}			
		}	
  		repaint();
	}
	
	// Sets the image2.
	/**	
	 * Das zweite Halbbild wird in seine Pixel aufgeteilt, um es sp�ter einzeln f�rben zu k�nnen.
	 * @param image2 zweite Halbbild
	 */
	protected void setImage2(BufferedImage image2){		
	  	intBufferImage2 = ((DataBufferInt)image2.getRaster().getDataBuffer()).getData();
	  		
	}	
		
	// RedBlue anaglyps.
	/**	 
	 * Die Rot-Blaue-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben wird das erste Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird das zweite Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Blaukanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Rot-Blau-Anaglyphen vereinigt. 
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf
	 */
	private void redBlueAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;
		
		int i = 0;		
		int rLeft = 0;
		int gLeft = 0;
		int bLeft = 0;
		int rRight = 0;
		int gRight = 0;
		int bRight = 0;
		
		int newRedLeft = 0;
		int newBlueRight = 0;
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) {   												
				if(intBufferImage1.length == intBufferImage2.length){
					rLeft = (intBufferImage1[i] >> 16) & 0xFF;	
					gLeft = ((intBufferImage1[i] >>  8) & 0xFF);	
					bLeft = intBufferImage1[i] & 0xFF;
					newRedLeft = (int)(red * rLeft + green * gLeft + blue * bLeft);		
					
					rRight = (intBufferImage2[i] >> 16) & 0xFF;
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;
					newBlueRight = (int)(red * rRight + green * gRight + blue * bRight);
									
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
	       							
					mergedIntBuffer[i++] = ((alpha << 24) | (newRedLeft << 16) | (0 << 8) | newBlueRight);
				}				
			}          
		}	
	}	
	
	// RedGreen anaglyps.
	/**	
	 * Die Rot-Gr�ne-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben wird das erste Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird das zweite Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Gr�nkanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Rot-Gr�n-Anaglyphen vereinigt. 
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf
	 */
	private void redGreenAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;		
		
		int i = 0;		
		int rLeft = 0;
		int gLeft = 0;
		int bLeft = 0;
		int rRight = 0;
		int gRight = 0;
		int bRight = 0;
		
		int newRedLeft = 0;
		int newGreenRight = 0;
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) {  
				if(intBufferImage1.length == intBufferImage2.length){
					rLeft = (intBufferImage1[i] >> 16) & 0xFF;	
					gLeft = ((intBufferImage1[i] >>  8) & 0xFF);	
					bLeft = intBufferImage1[i] & 0xFF;
					newRedLeft = (int)(red * rLeft + green * gLeft + blue * bLeft);		
					
					rRight = (intBufferImage2[i] >> 16) & 0xFF;
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;
	
					newGreenRight = (int)(red * rRight + green * gRight + blue * bRight);
									
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
	       							
					mergedIntBuffer[i++] = ((alpha << 24) | (newRedLeft << 16) | (newGreenRight << 8) | 0);
				}
			}          
		}	
	}	
	
	// Gray anaglyps.
	/**	 
	 * Die Graue-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben wird das erste Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird das zweite Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Blau- und Gr�nkanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Grau-Anaglyphen vereinigt. 
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf
	 */
	private void grayAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;		 
		
		int i = 0;		
		int rLeft = 0;
		int gLeft = 0;
		int bLeft = 0;
		int rRight = 0;
		int gRight = 0;
		int bRight = 0;
		
		int newRedLeft = 0;
		int newGreenRight = 0;
		int newBlueRight = 0;
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) {   
				if(intBufferImage1.length == intBufferImage2.length){

					rLeft = (intBufferImage1[i] >> 16) & 0xFF;	
					gLeft = ((intBufferImage1[i] >>  8) & 0xFF);	
					bLeft = intBufferImage1[i] & 0xFF;
					newRedLeft = (int)(red * rLeft + green * gLeft + blue * bLeft);		
				
					rRight = (intBufferImage2[i] >> 16) & 0xFF;
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;
					newGreenRight = (int)(red * rRight + green * gRight + blue * bRight);
					newBlueRight = (int)(red * rRight + green * gRight + blue * bRight);
									
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
				}			
				mergedIntBuffer[i++] = ((alpha << 24) | (newRedLeft << 16) | (newGreenRight << 8) | newBlueRight);		
				
			}          
		}	
	}
		
	// Color anaglyps.
	/**	
	 * Die Farbige-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben wird der Rotkanal des erste Halbbild ohne f�rben im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird der Blau- und Gr�nkanal des zweite Halbbild ohne f�rben im Blau- und Gr�nkanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Farben-Anaglyphen vereinigt.
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf
	 */
	private void colorAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;		 
		
		int i = 0;		
		int rLeft = 0;
		int gRight = 0;
		int bRight = 0;	
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) { 
				if(intBufferImage1.length == intBufferImage2.length){
					rLeft = (intBufferImage1[i] >> 16) & 0xFF;	
							
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;
													
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
	       							
					mergedIntBuffer[i++] = ((alpha << 24) | (rLeft << 16) | (gRight << 8) | bRight);
				}
			}          
		}	
	}
	
	// Half color anaglyps.
	/**	
	 * Die Halbfarbige-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben wird das erste Halbbild in Graustufen gef�rbt und die erzeugten Graustufen im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird der Blau- und Gr�nkanal des zweite Halbbild ohne f�rben im Blau- und Gr�nkanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Halbfarben-Anaglyphen vereinigt.
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf 
	 */
	private void halfColorAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;		 
		
		int i = 0;		
		int rLeft = 0;
		int gLeft = 0;
		int bLeft = 0;
		int gRight = 0;
		int bRight = 0;
		
		int newRedLeft = 0;
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) {  
				if(intBufferImage1.length == intBufferImage2.length){
					rLeft = (intBufferImage1[i] >> 16) & 0xFF;	
					gLeft = ((intBufferImage1[i] >>  8) & 0xFF);	
					bLeft = intBufferImage1[i] & 0xFF;
					newRedLeft = (int)(red * rLeft + green * gLeft + blue * bLeft);		
									
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;				
									
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
	       							
					mergedIntBuffer[i++] = ((alpha << 24) | (newRedLeft << 16) | (gRight << 8) | bRight);
				}
			}          
		}	
	}
	
	// Optimized anaglyps.
	/**	
	 * Die Optimierte-Anaglyphen F�rbemethode.
	 * Durch pixelweises f�rben werden die Graustufen des ersten Halbbildes nur aus blau und gr�n erzeugt, die erzeugten Graustufen werden im Rotkanal gespeichert.
	 * Durch pixelweises f�rben wird der Blau- und Gr�nkanal des zweite Halbbild ohne f�rben im Blau- und Gr�nkanal gespeichert.
	 * Die beiden gef�rbten Halbbilder werden zu einem Optimierten-Anaglyphen vereinigt.
	 * @param width Breite
	 * @param height H�he
	 * @param mergedIntBuffer nimmt die vereinigten Halbbildpixel auf
	 */
	private void optimizedAnaglyps(int width, int height, int[] mergedIntBuffer){
		int alpha = 0;		 
		
		int i = 0;			
		int gLeft = 0;
		int bLeft = 0;
		int gRight = 0;
		int bRight = 0;
		
		int newRedLeft = 0;	
		
		for (int j=0; j < height; j++) {
			for (int k=0; k < width; k++) { 		
				if(intBufferImage1.length == intBufferImage2.length){
					gLeft = ((intBufferImage1[i] >>  8) & 0xFF);	
					bLeft = intBufferImage1[i] & 0xFF;
					newRedLeft = (int)(0.837 * gLeft + 0.163 * bLeft);		
								
					gRight = ((intBufferImage2[i] >>  8) & 0xFF);
					bRight = intBufferImage2[i] & 0xFF;				
									
					alpha = (intBufferImage2[i] >> 24) & 0xFF; 				                        
	       							
					mergedIntBuffer[i++] = ((alpha << 24) | (newRedLeft << 16) | (gRight << 8) | bRight);
				}
			}          
		}	
	}
	
	// Stop rendering.
	/**	
	 * 
	 * Startet das on-screen Rendern der Klasse <code>AnaglyphCanvas3D</code> und stoppt das off-screen Rendern der Klasse <code>OffscreenCanvas3D</code>.
	 */
	private void stopRendering(){
		startRenderer();	
		canvas3DOfflineImage1.stopOffScreenRendering();
		canvas3DOfflineImage2.stopOffScreenRendering();
		refreshSizeTimer.cancel();
	}
	
	 // Start rendering.
	/**
	 * Stoppt das on-screen Rendern der Klasse <code>AnaglyphCanvas3D</code> und startet das off-screen Rendern der Klasse <code>OffscreenCanvas3D</code>.
	 */
	private void startRendering(){
		stopRenderer();
		canvas3DOfflineImage1.startOffScreenRendering();
		canvas3DOfflineImage2.startOffScreenRendering();
		refreshSizeTimer = new Timer("Refresh AnaglyphCanvas Size");
		refreshSizeTimer.schedule ( new TimerTask() {
			@Override
			public void run() {
				Dimension size = getSize();
				if(size!=null&&oldSize!=null&&(size.getWidth()!=oldSize.getWidth()||size.getHeight()!=oldSize.getHeight())) {
					canvas3DOfflineImage1.setOffScreenBuffer(size.width,size.height);
					canvas3DOfflineImage2.setOffScreenBuffer(size.width,size.height);
					oldSize = size;
				}
			}
		}, 100, 100);		
	}
	
	
	// Sets the anaglyph mode.
	/**
	 * Setzt den aktuellen Anaglyphenmodus.
	 * @param anaglyphMode Anglyphenmodus
	 */
	public void setAnaglyphMode(AnaglyphMode anaglyphMode){
		this.anaglyphMode = anaglyphMode;		
	}
	
	// Gets the anaglyph mode.
	/**
	 * Zeigt den aktuell eingestellten Anaglyphenmodus an.
	 * @return Anaglyphenmodus Anglyphenmodus
	 */
	public AnaglyphMode getAnaglyphMode(){
		return anaglyphMode;
	}
	
	// Sets the stereo mode.
	/**
	 * Setzt den aktuellen Stereomodus.
	 * @param stereoMode Stereomodus	
	 */
	public void setStereoMode(StereoMode stereoMode){		
		this.stereoMode = stereoMode;
		switch(stereoMode){
		case OFF:	
			if(!isRendererRunning())stopRendering();
			break;
		case ANAGLYPH:
			if(canvas3DOfflineImage1==null)
				initializeAnaglyphRendering();
			if(isRendererRunning())startRendering();					
			break;		
		case POLARIZED:
			if(!isRendererRunning())stopRendering();
			break;		
		case SHUTTER_GLASSES:
			if(!isRendererRunning())stopRendering();
			break;		
		}		
	}

	// Gets the stereo mode.
	/**	
	 * Zeigt den aktuell eingestellten Stereomodus an.
	 * @return stereoMode Stereomodus
	 */
	public StereoMode getStereoMode(){
		return stereoMode;
	}
	
	// Sets the stereo properties.
	/**	
	 * Setzt den Augenabstand und erm�glicht das Vertauschen der Augen.#
	 * Here is a bug. The eye distance will be refreshed in the image after clicking the canvas.
	 * 
	 * @param eyeDistance Augenabstand
	 * @param switchEyes Augentauschen
	 */
	public void setStereoProperties(double eyeDistance, boolean switchEyes){			
		double switchValue = (switchEyes?-1:1)*(eyeDistance/5000d);
		
		getView().getPhysicalBody().setLeftEyePosition(new Point3d(-switchValue, 0, 0));
		getView().getPhysicalBody().setRightEyePosition(new Point3d(switchValue, 0, 0));
	}
		
		
	/* (non-Javadoc)
	 * @see java.awt.Canvas#update(java.awt.Graphics)
	 */
	@Override
	public void update(Graphics g) {
		//with the standard update the canvas gets cleared
		//and then paint() is called. This leads to flickering
		//and is overridden here
		paint(g);
	}

	/* (non-Javadoc)
	 * @see javax.media.j3d.Canvas3D#paint(java.awt.Graphics)
	 */
	@Override
	public void paint(Graphics g) {
		super.paint(g);
		
		if (mergedImage!=null && isEnabled())
			g.drawImage(mergedImage, 0, 0, null);
	}		
	
	//TODO: to get rid off
	public void setTheParent(JComponent parent) {
        this.parent = parent;
	}
	
	public void setTransparencySortingPolicy(int transparencySortGeometry) {
        getView().setTransparencySortingPolicy(transparencySortGeometry);
        transparencySort = transparencySortGeometry;           
	}
	
	/**
	 * Should be called when the canvas shall be closed. It will free all resources and
	 * close the OffscreeenCanvas. Otherwise a memory leak will occur. 
	 */
	public void close() {
		intBufferImage1 = null;
		intBufferImage2 = null;
		parent = null;
		if(canvas3DOfflineImage1!=null)
			canvas3DOfflineImage1.close();
		canvas3DOfflineImage1 = null;
		if(canvas3DOfflineImage2!=null)
			canvas3DOfflineImage2.close();
		canvas3DOfflineImage2 = null;
		if(refreshSizeTimer!=null)
			refreshSizeTimer.cancel();
		refreshSizeTimer = null;
	}
	
}