//==============================================================================
//
//		Copyright 2009 IPK Gatersleben, AG Plant Bioinformatics
//		Developed by K. Pucknat, H. Rohn
//
//		Licensed under the Apache License, Version 2.0 (the "License");
//		you may not use this file except in compliance with the License.
//		You may obtain a copy of the License at
//
//			http://www.apache.org/licenses/LICENSE-2.0
//
//		Unless required by applicable law or agreed to in writing, software
//		distributed under the License is distributed on an "AS IS" BASIS,
//		WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//		See the License for the specific language governing permissions and
//		limitations under the License. 
//
//		Portions of this software is based in part on the CyberVRML97 package written by Satoshi Konno.
//
//==============================================================================
package anaglyphcanvas3d;

import java.awt.GraphicsConfiguration;
import java.awt.image.BufferedImage;

import javax.media.j3d.Canvas3D;
import javax.media.j3d.ImageComponent;
import javax.media.j3d.ImageComponent2D;
import javax.media.j3d.RestrictedAccessException;
import javax.media.j3d.View;

import com.sun.j3d.exp.swing.impl.AutoOffScreenCanvas3D;

// TODO: Auto-generated Javadoc
// The Class Canvas3DOffScreen.
/** 
 * Die Klasse <code>OffscreenCanvas3D</code> wird ausschlie�lich zum off-screen Rendern der Halbbilder im Anaglyphenmodus benutzt. 
 * @see Canvas3D
 * @see AnaglyphCanvas3D
 */
class OffScreenCanvas3D extends Canvas3D implements AutoOffScreenCanvas3D{
	//The Constant serialVersionUID.
	/** Die Konstante serialVersionUID. */
	private static final long serialVersionUID = 1L;	
	 
	/** 
	 * Referenz auf die Klasse <code>AnaglyphCanvas3D</code>, damit die Halbbilder �bergeben werden k�nnen.
	 * */
	private AnaglyphCanvas3D canvas3DOnline;	
	
	//The off. 
	/** 
	 * Die Variable wird benutzt, um zu verhindern, dass off-screen gerendert wird, obwohl der Stereoskopie-Modus deaktiviert ist.
	 * @see stopOffscreenRendering
	 * @see startOffscreenRendering
	 * 
	 * */
	private boolean off = false;
	
	//Instantiates a new canvas3 d off screen.
	/**
	 * Konstruiert und initialisiert ein neues <code>OffscreenCanvas3D</code> Objekt.
	 * @param graphicsConf Grafik Konfiguration (z.B. <code>SimpleUniverse.getPreferredConfiguration()</code>)		 
	 * @param eye Sicht f�r das Linke oder Rechte Auge des Betrachters
	 * @see GraphicsConfiguration	
	 */
	public OffScreenCanvas3D(GraphicsConfiguration graphicsConf, int eye) {
		super(graphicsConf,true);
		setMonoscopicViewPolicy(eye);	  
	}
	
	//Sets the canvas3 d online ref.
	/**
	 *  Referenz auf die Klasse <code>AnaglyphCanvas3D</code> damit die Halbbilder �bergeben werden k�nnen.
	 * @param canvas3DOnline Referenz auf <code>AnaglyphCanvas3D</code> Objekt
	 */
	void setCanvas3DOnlineRef(AnaglyphCanvas3D canvas3DOnline){
		this.canvas3DOnline = canvas3DOnline; 
	}
		
	//Sets the off screen buffer.
	/**
	 * Setzt den off-screen Buffer und veranlasst das automatische Rendern, insofern der Stereoskopie-Modus nicht deaktiviert ist.
	 * @param width Breite
	 * @param height H�he
	 */
	void setOffScreenBuffer(int width, int height){	
		this.stopRenderer(); 
		
		if(width <= 0 || height <= 0){
			width = 1;
			height = 1;
		}
		
		BufferedImage renderImage = new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );		
		ImageComponent2D image = new ImageComponent2D(ImageComponent.FORMAT_RGB, renderImage, true, false);
		image.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
  		image.setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);
  			
		this.waitForOffScreenRendering();  		
		
		try {      				
			this.setOffScreenBuffer(image);  			
		}catch (RestrictedAccessException e) {         
			System.out.println("RestrictedAccessException");
		}
  	
  		if(!off){	
			this.startRenderer();
  		}
	}
		
	//Stop off screen rendering.
	/**	 
	 * Stoppt das off-screen Rendern.
	 */
	protected void stopOffScreenRendering(){
		stopRenderer();
		off = true;
	}
	
	// * Start off screen rendering.
	/**	
	 * Startet das off-screen Rendern.
	 */
	protected void startOffScreenRendering(){
		startRenderer();
		off = false;
	}
	
	/* (non-Javadoc)
	 * @see javax.media.j3d.Canvas3D#postSwap()
	 */
	@Override
	public void postSwap() {		
		super.postSwap();		
		if(getMonoscopicViewPolicy()==View.LEFT_EYE_VIEW&&!off){		
			canvas3DOnline.setImage1(getOffScreenBuffer().getImage());
		}
		if(getMonoscopicViewPolicy()==View.RIGHT_EYE_VIEW&&!off){			
			canvas3DOnline.setImage2(getOffScreenBuffer().getImage());
		}		
	}	
	
	public void close() {
		stopOffScreenRendering();
		canvas3DOnline = null;
	}
	
}
